/*++
 *
 * Copyright (c) 2004-2006 Intel Corporation - All Rights Reserved
 *
 * This software program is licensed subject to the BSD License, 
 * available at http://www.opensource.org/licenses/bsd-license.html
 *
 * Abstract: The main routine
 *
 * Source code modified by EnterpriseDB Corporation, to suit PostgreSQL's code.
 * 
 --*/

#include "crc_sb8.h"

/**
 *
 * Routine Description:
 *
 * Computes the CRC32c checksum for the specified buffer using the slicing by 8 
 * algorithm over 64 bit quantities.                      
 *
 * Arguments:
 *
 *		p_running_crc - pointer to the initial or final remainder value 
 *						used in CRC computations. It should be set to 
 *						non-NULL if the mode argument is equal to CONT or END
 *		p_buf - the packet buffer where crc computations are being performed
 *		length - the length of p_buf in bytes
 *		init_bytes - the number of initial bytes that need to be procesed before
 *					 aligning p_buf to multiples of 4 bytes
 *		mode - can be any of the following: BEGIN, CONT, END, BODY, ALIGN 
 *
 * Return value:
 *		
 *		The computed CRC32c value
 */

pg_crc32
crc32c_sb8_64_bit(
	pg_crc32* p_running_crc,
    const uint8*	p_buf,
    const uint32 length,
	CRC_SB8_MODE mode)
{
	uint32 li;
	uint32 crc, term1, term2;
	uint32 running_length;
	uint32 end_bytes;
	uint32 init_bytes;

	if(mode ==  CRC_SB8_MODE_CONT)
		crc = *p_running_crc;
	else	
		crc = CRC32C_INIT_REFLECTED;

	if (length)
	{
		init_bytes = ((uint32)p_buf)%4;
		if (init_bytes)
			init_bytes = 4 - init_bytes;
	}
	else
		init_bytes = 0;

	running_length = ((length - init_bytes)/8)*8;
	end_bytes = length - init_bytes - running_length; 

	for(li=0; li < init_bytes; li++) 
		crc = crc_tableil8_o32[(crc ^ *p_buf++) & 0x000000FF] ^ (crc >> 8);	
	for(li=0; li < running_length/8; li++) 
	{
		crc ^= *(uint32 *)p_buf;
		p_buf += 4;
		term1 = crc_tableil8_o88[crc & 0x000000FF] ^
				crc_tableil8_o80[(crc >> 8) & 0x000000FF];
		term2 = crc >> 16;
		crc = term1 ^
			  crc_tableil8_o72[term2 & 0x000000FF] ^ 
			  crc_tableil8_o64[(term2 >> 8) & 0x000000FF];
		term1 = crc_tableil8_o56[(*(uint32 *)p_buf) & 0x000000FF] ^
				crc_tableil8_o48[((*(uint32 *)p_buf) >> 8) & 0x000000FF];
		
		term2 = (*(uint32 *)p_buf) >> 16;
		crc =	crc ^ 
				term1 ^		
				crc_tableil8_o40[term2  & 0x000000FF] ^	
				crc_tableil8_o32[(term2 >> 8) & 0x000000FF];	
		p_buf += 4;
	}
	for(li=0; li < end_bytes; li++) 
		crc = crc_tableil8_o32[(crc ^ *p_buf++) & 0x000000FF] ^ (crc >> 8);
	if((mode == CRC_SB8_MODE_BEGIN) || (mode ==  CRC_SB8_MODE_CONT))
		return crc;		
    return crc ^ XOROT;	
}
